#!/usr/bin/env python3.1

import ply.lex
import ply.yacc
from operator import add, sub, mul, floordiv, mod

VERSION = '0.1'

class SexpException(Exception): pass

tokens = ('INT', 'DEF', 'ADD', 'SUB', 'MUL', 'DIV', 'MOD',  'LPAR', 'RPAR', 'ID',)

t_ignore = ' \t'

t_DEF   = r'def'
t_ADD   = r'[+]'
t_SUB   = r'-'
t_MUL   = r'[*]'
t_DIV   = r'div'
t_MOD   = r'mod'
t_LPAR  = r'[(]'
t_RPAR  = r'[)]'

keywords = {
    'def': 'DEF',
    'div': 'DIV',
    'mod': 'MOD'    
}    
    
def t_ID(t):
    r'[a-zA-Z]\w*'
    t.type = keywords.get(t.value, 'ID')
    return t    

def t_INT(t):
    r'-?\d+'
    t.value = int(t.value)
    return t
    
def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")
    
def t_error(t):
    raise SexpException("Illegal character '{0}'".format(t.value[0]))

symbol_table = {}

def p_statement_expr(p):
    'statement : exp'
    p[0] = p[1]

def p_defexp(p):
    'exp : LPAR DEF ID exp RPAR'
    symbol_table[p[3]] = p[4]
    p[0] = ''
    
def p_idexp(p):
    'exp : ID'
    p[0] = symbol_table.get(p[1], 0)
    
def p_intexp(p):
    'exp : INT'
    p[0] = p[1]

operator = {'+': add, '-': sub, '*': mul, 'div': floordiv, 'mod': mod}

def p_arithexp(p):
    'exp : LPAR op exp exp RPAR'
    p[0] = operator[p[2]](p[3], p[4])
    
def p_operator(p):
    '''op : ADD 
          | SUB 
          | MUL 
          | DIV 
          | MOD'''
    p[0] = p[1]

def p_error(p):
    if p:
        raise SexpException("Syntax error at '{0}'".format(p.value))
    else:
        raise SexpException("Syntax error")

# Build the lexer and parser
ply.lex.lex()
ply.yacc.yacc()

def interpret(s, table):
    global symbol_table
    symbol_table = table
    return ply.yacc.parse(s)    

if __name__ == '__main__':
    table = {}
    print('S-Expression Language REPL, version {0}'.format(VERSION))
    print('Quit the REPL with CONTROL-D.')
    while True:
        try:
            user_input = input('> ')
            result = interpret(user_input, table)
            if result != '':
                print(result)
        except EOFError:
            print('Hasta la vista, baby.')
            break
        except SexpException as e:
            print(e)

